# -*- coding:utf-8 -*-
# PROJECT_NAME : laravel_validator
# FILE_NAME    : 
# AUTHOR       : younger shen
import copy
from importlib import import_module
from django.core.exceptions import ValidationError
from .exceptions import InvalidRuleNameError
from .exceptions import InvalidValidateDataError
from .utils import format_args_split, check_errors, error_message_generate
from .rules import WITH_PARAMETERS_VALIDATOR


class BaseValidator(type):
    """
        Metaclass for all validators
    """

    def __init__(cls, name, bases, dct):
        new_attrs = dict()
        error_list = dict()
        error_list_ext = dict()
        if name is not 'Validator':
            for k in dct:
                if k != '__module__' and k != '__main__':
                    pattern = dct.get(k, None)
                    if pattern and not callable(pattern):
                        new_attrs.update(**{k: pattern})
                        error_list.update(**{k: {}})

            setattr(cls, 'validate_data', new_attrs)
            setattr(cls, 'error_list', error_list)
            setattr(cls, 'error_list_ext', error_list_ext)

        super(BaseValidator, cls).__init__(name, bases, dct)

    def __new__(mcs, name, bases, dct):
        return type.__new__(mcs, name, bases, dct)


class Validator(object):
    """
        base class for all Validator rules
    """
    __metaclass__ = BaseValidator

    def __init__(self, data, message=None):
        if not data:
            raise InvalidValidateDataError()

        self.data = data
        self.message = message
        self.validate_flag = True
        self.error_list = copy.deepcopy(self.error_list)
        self.error_list_ext = copy.deepcopy(self.error_list_ext)

    def add_error(self, error):
        self.error_list_ext.update(error)

    def get(self, item):
        return self.data.get(item, None)

    def fails(self):
        validate_data = getattr(self, 'validate_data')
        for k in validate_data:
            rules = validate_data.get(k, None)
            if not rules:
                raise InvalidValidateDataError()
            else:
                rules_list = rules.split('|')
                for rule in rules_list:
                    rule_origin = rule
                    rule = rule.split(':')[0]
                    if rule is None or rule == '':
                        raise InvalidRuleNameError(rule=rule)

                    rule_validator = import_module('.rules', package='django_laravel_validator')

                    try:
                        regex = getattr(rule_validator, rule.upper())
                    except AttributeError:
                        raise InvalidRuleNameError(rule=rule)

                    if rule.upper() in WITH_PARAMETERS_VALIDATOR:
                        rule_args = format_args_split(rule_origin)
                        regex = regex(rule_args)
                    else:
                        regex = regex()
                    try:
                        setattr(regex, 'validator_instance', self)
                        regex(self.data.get(k, None))
                    except ValidationError as e:
                        error_message = error_message_generate(k, rule, self.message, e)
                        error_dict = self.error_list.get(k)
                        error_dict.update({rule: error_message})
                        self.error_list.get(k).update(error_dict)
                        self.validate_flag = False

        check = getattr(self, 'check', None)
        if check and callable(check) and self.validate_flag:
            check()

        return check_errors(self.error_list, self.error_list_ext)

    def errors(self):
        self.error_list.update(extra=self.error_list_ext)
        return self.error_list



# -*- coding:utf-8 -*-
# PROJECT_NAME : laravel_validator
# FILE_NAME    : 
# AUTHOR       : younger shen

from django_laravel_validator.exceptions import InvalidMinValidatorParameterError
from django_laravel_validator.utils import check_errors
from django_laravel_validator.validator import Validator
from .messages import *


class TestValidator1(Validator):
    username = 'required'
    password = 'required'

    def check(self):
        username = self.data.get('username')
        password = self.data.get('password')

        if username != 'youngershen':
            self.error_list.get('username').update(dict(wrong_user=u'you got a wrong user'))

        if password != '123456':
            self.error_list.get('password').update(dict(wrong_pass=u'you got a wrong pass'))


class TestValidator2(Validator):
    username = 'required|min:8|numeric'
    password = 'required|min:8|numeric'

    def check(self):
        print ' i am checking '


class TestValidator3(Validator):

    username = 'required|min:sdf|numeric|sss'
    password = 'required|min:sss|numeric'


class TestValidator4(Validator):
    username = 'required|min:8|numeric'
    password = 'required|min:2|numeric'


class TestValidator5(Validator):
    username = 'required|min:8|max:16|alpha'
    password = 'required|range:2,5|alpha'


class TestValidator6(Validator):
    username = 'required|length:5|alpha'
    password = 'required|range:2,5'
    remember_me = 'accepted'
    email = 'required|email'
    ip = 'ip'
    is_superuser = 'bool'
    phone = 'required|regex:^[0-9]{5,10}$'


class MatchValidatorTestValidator(Validator):
    password = 'required'
    password_confirm = 'required|match:password'


class UniqueValidatorTestValidator(Validator):
    email = 'required|unique:account.User,email'


def test_match_validator():
    data = dict(password='123', password_confirm='123')
    validator = MatchValidatorTestValidator(data)
    ret = validator.fails()
    assert ret


def test_validator():
    data = dict(username='youngershen', password='123456')
    validator = TestValidator1(data)
    ret = validator.fails()
    error_list = validator.errors()

    assert ret is True
    assert check_errors(error_list, None)

    data = dict(username='123', password='123')

    validator = TestValidator1(data)
    ret = validator.fails()
    error_list = validator.errors()

    assert ret is False
    assert check_errors(error_list, None) is False


def test_validator2():
    data = dict(username='123456789', password='123456789')
    validator = TestValidator2(data)
    ret = validator.fails()
    error_list = validator.errors()

    assert ret is True
    assert check_errors(error_list, None) is True

    data = dict(username='', password='wwwsf')
    validator = TestValidator2(data)
    ret = validator.fails()
    error_list = validator.errors()

    assert ret is False
    assert error_list.get('username').get('required') == REQUIRED_MESSAGE
    assert error_list.get('username').get('min') == MIN_MESSAGE.format(min=8)
    assert error_list.get('username').get('numeric') == NUMERIC_MESSAGE
    assert error_list.get('password').get('min') == MIN_MESSAGE.format(min=8)
    assert error_list.get('password').get('numeric') == NUMERIC_MESSAGE


def test_validator3():
    data = dict(username='tinker', password='i am tinker')
    validator = TestValidator3(data)
    try:
        ret = validator.fails()
    except InvalidMinValidatorParameterError as e:
        pass


def test_validator4():
    data = dict(username='', password='')
    message = {'username.required': 'haha', 'password.required': 'heihei', 'username.min': 'min'}

    validator = TestValidator4(data, message=message)

    ret = validator.fails()
    error_list = validator.errors()
    assert ret is False
    assert error_list.get('username').get('required') == 'haha'
    assert error_list.get('password').get('required') == 'heihei'
    assert error_list.get('username').get('min') == 'min'
    assert error_list.get('username').get('numeric') == NUMERIC_MESSAGE


def test_validator5():
    data = dict(username='', password='')
    validator = TestValidator5(data)
    ret = validator.fails()
    error_list = validator.errors()

    assert ret is False

    assert error_list.get('username').get('required') == REQUIRED_MESSAGE
    assert error_list.get('username').get('min') == MIN_MESSAGE.format(min=8)
    assert error_list.get('username').get('max') == MAX_MESSAGE.format(max=16)
    assert error_list.get('username').get('alpha') == ALPHA_MESSAGE

    assert error_list.get('password').get('required') == REQUIRED_MESSAGE
    assert error_list.get('password').get('range') == RANGE_MESSAGE.format(min=2, max=5)
    assert error_list.get('password').get('alpha') == ALPHA_MESSAGE


def test_validator6():
    data = dict(username='abcef', password='abc', remember_me='1', email='younger.x.shen@gmail.com', ip='192.168.1.1', is_superuser='1', phone='123456')
    validator = TestValidator6(data)
    ret = validator.fails()
    error_list = validator.errors()

    assert not error_list.get('username')
    assert not error_list.get('password')
    assert not error_list.get('remember_me')
    assert not error_list.get('email')
    assert not error_list.get('ip')
    assert not error_list.get('is_superuser')


def test_validator7():
    data = dict(username='youngershen', email='root@root.com')
    validator = UniqueValidatorTestValidator(data)
    ret = validator.fails()
    assert ret is True



# -*- coding:utf-8 -*-
# PROJECT_NAME : laravel_validator
# FILE_NAME    : 
# AUTHOR       : younger shen
from importlib import import_module
import re
from django.core.exceptions import ValidationError, FieldError
from django.core.validators import RegexValidator
from .regex import RE_NUMBERIC
from .regex import RE_ALPHA
from .regex import RE_EMAIL
from .regex import RE_IP_ADDRESS
from .exceptions import InvalidMinValidatorParameterError
from .exceptions import InvalidMaxValidatorParameterError
from .exceptions import InvalidRangeValidatorParameterError
from .exceptions import InvalidLengthValidatorParameterError
from .exceptions import InvalidRegexValidatorParameterError
from .exceptions import InvalidMatchValidatorParameterError
from .exceptions import InvalidUniqueValidatorParameterError
from .messages import REQUIRED_MESSAGE
from .messages import NUMERIC_MESSAGE
from .messages import MIN_MESSAGE
from .messages import MAX_MESSAGE
from .messages import RANGE_MESSAGE
from .messages import LENGTH_MESSAGE
from .messages import ACCEPTED_MESSAGE
from .messages import ALPHA_MESSAGE
from .messages import EMAIL_MESSAGE
from .messages import IP_ADDRESS_MESSAGE
from .messages import BOOLEAN_MESSAGE
from .messages import REGEX_MESSAGE
from .messages import MATCH_MESSAGE
from .messages import UNIQUE_MESSAGE
from .messages import UNIQUE_UNKNOW_APP_NAME
from .messages import UNIQUE_UNKNOW_MODEL_NAME
from .messages import UNIQUE_UNKNOW_MODEL_FIELD
WITH_PARAMETERS_VALIDATOR = ['MIN', 'MAX', 'RANGE', 'LENGTH', 'REGEX', 'MATCH', 'UNIQUE']


class BaseValidator(RegexValidator):
    pass


class RequiredValidator(BaseValidator):

    def __init__(self, **kwargs):
        super(RequiredValidator, self).__init__(**kwargs)
        self.code = 'required'
        self.message = REQUIRED_MESSAGE

    def __call__(self, value=None):
        if value is None or value == '':
            raise ValidationError(message=self.message, code=self.code)

        if value and value.strip() == '':
            raise ValidationError(message=self.message, code=self.code)


class NumericValidator(BaseValidator):

    def __init__(self, **kwargs):
        super(NumericValidator, self).__init__(**kwargs)
        self.code = 'numeric'
        self.message = NUMERIC_MESSAGE

    def __call__(self, value=None):
        if not re.match(RE_NUMBERIC, value):
            raise ValidationError(message=self.message, code=self.code)


class MinValidator(BaseValidator):

    def __init__(self, args=None, **kwargs):
        super(MinValidator, self).__init__(**kwargs)

        if not args or len(args) != 1 or args[0] is None or args[0] == '':
            raise InvalidMinValidatorParameterError()

        if not re.match(RE_NUMBERIC, args[0]):
            raise InvalidMinValidatorParameterError()

        self.min = int(args[0])
        self.code = 'min'
        self.message = MIN_MESSAGE.format(min=self.min)

    def __call__(self, value=None):

        if value is None or value == '':
            raise ValidationError(message=self.message, code=self.code)

        if len(value) < self.min:
            raise ValidationError(message=self.message, code=self.code)


class MaxValidator(BaseValidator):

    def __init__(self, args=None, **kwargs):
        super(MaxValidator, self).__init__(**kwargs)

        if not args or len(args) != 1 or args[0] is None or args[0] == '':
            raise InvalidMaxValidatorParameterError()

        if not re.match(RE_NUMBERIC, args[0]):
            raise InvalidMaxValidatorParameterError()

        self.max = int(args[0])
        self.code = 'max'
        self.message = MAX_MESSAGE.format(max=self.max)

    def __call__(self, value=None):
        if value is None or value == '':
            raise ValidationError(message=self.message, code=self.code)

        if len(value) > self.max:
            raise ValidationError(message=self.message, code=self.code)


class LengthValidator(BaseValidator):

    def __init__(self, args=None, **kwargs):
        super(LengthValidator, self).__init__(**kwargs)

        if not args or len(args) != 1:
            raise InvalidLengthValidatorParameterError()

        if not re.match(RE_NUMBERIC, args[0]):
            raise InvalidLengthValidatorParameterError()

        self.length = int(args[0])
        self.code = 'length'
        self.message = LENGTH_MESSAGE.format(length=self.length)

    def __call__(self, value=None):
        if value is None or value == '':
            raise ValidationError(message=self.message, code=self.code)

        if not len(value) == self.length:
            raise ValidationError(message=self.message, code=self.code)


class RangeValidator(BaseValidator):

    def __init__(self, args=None, **kwargs):
        super(RangeValidator, self).__init__(**kwargs)

        if not args or len(args) != 2:
            raise InvalidRangeValidatorParameterError()

        if not re.match(RE_NUMBERIC, args[0]) or not re.match(RE_NUMBERIC, args[1]):
            raise InvalidMinValidatorParameterError()

        self.min = int(args[0])
        self.max = int(args[1])

        if self.min > self.max:
            self.min, self.max = self.max, self.min

        self.code = 'range'
        self.message = RANGE_MESSAGE.format(min=self.min, max=self.max)

    def __call__(self, value=None):
        if value is None or value == '':
            raise ValidationError(self.message, code=self.code)

        value_length = len(value)

        if self.min == self.max:
            if not value_length == self.max:
                raise ValidationError(message=self.message, code=self.code)

        if self.min > value_length or self.max < value_length:
            raise ValidationError(message=self.message, code=self.code)


class AcceptedValidator(BaseValidator):

    def __init__(self, **kwargs):
        super(AcceptedValidator, self).__init__(**kwargs)

        self.code = 'accepted'
        self.message = ACCEPTED_MESSAGE

    def __call__(self, value=None):
        if value == 'yes' or value == 'on' or value == '1':
            pass
        else:
            raise ValidationError(message=self.message, code=self.code)


class AlphaValidator(BaseValidator):

    def __init__(self, **kwargs):
        super(AlphaValidator, self).__init__(**kwargs)

        self.code = 'alpha'
        self.message = ALPHA_MESSAGE

    def __call__(self, value=None):

        if not re.match(RE_ALPHA, value):
            raise ValidationError(message=self.message, code=self.code)


class EmailValidator(BaseValidator):

    def __init__(self, **kwargs):
        super(EmailValidator, self).__init__(**kwargs)
        self.code = 'email'
        self.message = EMAIL_MESSAGE

    def __call__(self, value=None):
        if not re.match(RE_EMAIL, value):
            raise ValidationError(message=self.message, code=self.code)


class IPAddressValidator(BaseValidator):

    def __init__(self, **kwargs):
        super(IPAddressValidator, self).__init__(**kwargs)
        self.code = 'ip'
        self.message = IP_ADDRESS_MESSAGE

    def __call__(self, value=None):
        if not re.match(RE_IP_ADDRESS, value):
            raise ValidationError(message=self.message, code=self.code)


class BooleanValidator(BaseValidator):

    def __init__(self, **kwargs):
        super(BooleanValidator, self).__init__(**kwargs)
        self.code = 'boolean'
        self.message = BOOLEAN_MESSAGE

    def __call__(self, value=None):
        if value == 1 or value == 0 or value == '1' or value == '0' or value == 'True' or value is True or value == 'False' or value is False:
            pass
        else:
            raise ValidationError(message=self.message, code=self.code)


class RegexValidator(BaseValidator):

    def __init__(self, args, **kwargs):
        super(RegexValidator, self).__init__(**kwargs)
        self.code = 'regex'

        if not args:
            raise InvalidRegexValidatorParameterError()
        regex = ''
        for s in args:
            regex += str(s)
            regex += ','

        self.regex = regex[:-1]
        self.message = REGEX_MESSAGE.format(regex=self.regex)

    def __call__(self, value=None):
        if not re.match(self.regex, value):
            raise ValidationError(message=self.message, code=self.code)


class MatchValidator(BaseValidator):

    def __init__(self, args, **kwargs):
        super(MatchValidator, self).__init__(**kwargs)
        self.code = 'match'

        if not args or len(args) != 1:
            raise InvalidMatchValidatorParameterError()

        self.match = args[0]
        self.message = MATCH_MESSAGE.format(match=self.match)

    def __call__(self, value=None):
        instance = getattr(self, 'validator_instance')
        data = getattr(instance, 'data', None)
        match_value = data.get(self.match, None)

        if not match_value:
            raise InvalidMatchValidatorParameterError()
        else:
            if not value == match_value:
                raise ValidationError(message=self.message, code=self.code)


class UniqueValidator(BaseValidator):

    def __init__(self, args, **kwargs):
        super(UniqueValidator, self).__init__(**kwargs)
        self.code = 'unique'
        if not args or len(args) != 2:
            raise InvalidUniqueValidatorParameterError()

        if args[0] == '' or args[1] == '':
            raise InvalidUniqueValidatorParameterError()

        temp = args[0].split('.')
        temp.insert(1, 'models')
        module_name = '.'.join(temp[:2])
        try:
            module = import_module(module_name)
        except ImportError:
            message = UNIQUE_UNKNOW_APP_NAME.format(appname=temp[0])
            raise InvalidUniqueValidatorParameterError(message)

        klass = getattr(module, temp[2], None)
        if not klass:
            message = UNIQUE_UNKNOW_MODEL_NAME.format(appname=temp[0], modelname=temp[2])
            raise InvalidUniqueValidatorParameterError(message)
        self.klass = klass
        self.field = args[1]
        self.message = UNIQUE_MESSAGE.format(field=self.field, model=temp[2])

    def __call__(self, value=None):
        try:
            self.klass.objects.get(**{self.field: value})
        except FieldError:
            message = UNIQUE_UNKNOW_MODEL_FIELD.format(fieldname=self.field)
            raise InvalidUniqueValidatorParameterError(message=message)
        except self.klass.DoesNotExist:
            return
        else:
            raise ValidationError(code=self.code, message=self.message)


REQUIRED = RequiredValidator
NUMERIC = NumericValidator
MIN = MinValidator
MAX = MaxValidator
RANGE = RangeValidator
LENGTH = LengthValidator
ACCEPTED = AcceptedValidator
ALPHA = AlphaValidator
EMAIL = EmailValidator
IP = IPAddressValidator
BOOL = BooleanValidator
REGEX = RegexValidator
MATCH = MatchValidator
UNIQUE = UniqueValidator


# -*- coding:utf-8 -*-
# PROJECT_NAME : invoker
# FILE_NAME    : 
# AUTHOR       : younger shen


class BaseValidatorException(Exception):
    pass


class InvalidValidateDataError(BaseValidatorException):

    def __init__(self, message=None):
        self.message = message if message else 'invalid validate data error'

    def __str__(self):
        return self.message


class InvalidMinValidatorParameterError(BaseValidatorException):

    def __init__(self, message=None):
        self.message = message if message else 'invalid MIN validator parameter error'

    def __str__(self):
        return self.message


class InvalidMaxValidatorParameterError(BaseValidatorException):

    def __init__(self, message=None):
        self.message = message if message else 'invalid MAX validator parameter error'

    def __str__(self):
        return self.message


class InvalidRangeValidatorParameterError(BaseValidatorException):

    def __init__(self, message=None):
        self.message = message if message else 'invalid RANGE validator parameter error'

    def __str__(self):
        return self.message


class InvalidLengthValidatorParameterError(BaseValidatorException):

    def __init__(self, message=None):
        self.mesage = message if message else 'invalid LENGTH validator parameter error'

    def __str__(self):
        return self.message


class InvalidAcceptedValidatorParameterError(BaseValidatorException):

    def __init__(self, message=None):
        self.message = message if message else 'invalid ACCEPTED validator parameter error'

    def __str__(self):
        return self.message


class InvalidActiveURLValidatorParameterError(BaseValidatorException):

    def __init__(self, message=None):
        self.message = message if message else 'invalid active_url validator parameter error'

    def __str__(self):
        return self.message


class InvalidAlphaValidatorParameterError(BaseValidatorException):

    def __init__(self, message=None):
        self.message = message if message else 'invalid alpha validator parameter error'

    def __str__(self):
        return self.message


class InvalidRegexValidatorParameterError(BaseValidatorException):

    def __init__(self, message=None):
        self.message = message if message else 'invalid regex validator parameter error'

    def __str__(self):
        return self.message


class InvalidDataError(BaseValidatorException):

    def __init__(self, message=None):
        self.message = message if message else 'invalid data error'

    def __str__(self):
        return self.message


class InvalidRuleNameError(BaseValidatorException):

    def __init__(self, message=None, rule=None):
        self.message = message if message else 'invalid rule name error'
        if rule:
            self.message = self.message + ': ' + rule

    def __str__(self):
        return self.message


class InvalidMatchValidatorParameterError(BaseValidatorException):

    def __init__(self, message=None):
        self.message = message if message else 'invalid match validator parameter error'

    def __str__(self):
        return self.message


class InvalidValidateDataError(BaseException):

    def __init__(self, message=None):
        self.message = message if message else 'invalid validate data error'

    def __str__(self):
        return self.message


class InvalidUniqueValidatorParameterError(Exception):

    def __init__(self, message=None):
        self.message = message if message else 'invalid unique validator parameter error'

    def __str__(self):
        return self.message




# -*- coding:utf-8 -*-
# PROJECT_NAME : laravel_validator
# FILE_NAME    : 
# AUTHOR       : younger shen


def format_args_split(format_str):
    if ':' in format_str:
        args = format_str.split(':')[1].split(',')
        args = map(lambda a: a.strip(), args)
        return args
    else:
        return None


def check_errors(error_list, error_list_ext):

    for key in error_list:
        value = error_list.get(key)
        if len(value.keys()) > 0:
            return False

    if error_list_ext and len(error_list_ext.keys()) > 0:
        return False

    return True


def error_message_generate(field, rule, message, error):
    if not message:
        return error[0]
    else:
        message_key = field + '.' + rule
        message_str = message.get(message_key, None)
        if message_str:
            return message_str
        else:
            return error[0]


# -*- coding:utf-8 -*-
# PROJECT_NAME : invoker
# FILE_NAME    : 
# AUTHOR       : younger shen

REQUIRED_MESSAGE = u'The field under validation must be present in the input data.'
NUMERIC_MESSAGE = u'The field under validation must have a numeric value.'
MIN_MESSAGE = u'The field under validation must have a minimum value : {min}.'
MAX_MESSAGE = u'The field under validation must be less than or equal to a maximum value : {max}'
LENGTH_MESSAGE = u'The field under validation must have a size of {length}'
RANGE_MESSAGE = u'The field under validation must have a size between the given min and max of {min} to {max}'
ACCEPTED_MESSAGE = u'The field under validation must be yes, on, or 1.'
ALPHA_MESSAGE = u'The field under validation must be entirely alphabetic characters'
EMAIL_MESSAGE = u'The field under validation must be formatted as an e-mail address.'
IP_ADDRESS_MESSAGE = u'The field under validation must be formatted as an IP address.'
BOOLEAN_MESSAGE = u'The field under validation must be able to be cast as a boolean. Accepted input are true, false, 1, 0, "1" and "0".'
REGEX_MESSAGE = u'The field under validation must match the given regular expression : {regex} .'
MATCH_MESSAGE = u'the field must match : {match}. '
UNIQUE_MESSAGE = u'the {field} field must be unique of table {model}.'
# error messages

UNIQUE_UNKNOW_APP_NAME = u'unknow app name : {appname}'
UNIQUE_UNKNOW_MODEL_NAME = u'unknow model name : {appname}.models.{modelname}'
UNIQUE_UNKNOW_MODEL_FIELD = u'unknow model field name : {fieldname}'


from setuptools import setup, find_packages
import sys, os

version = '0.0.1a6'

setup(name='django-laravel-validator',
      version=version,
      description="a data validator for django inspired the validator of laravel framework",
      long_description="""\
      a data validator for django inspired the validator of laravel framework
      easy to use easy to validate every thing.
      """,
      classifiers=['Development Status :: 2 - Pre-Alpha',
                   'Environment :: Web Environment',
                   'Framework :: Django',
                   'Intended Audience :: Developers',
                   'License :: OSI Approved :: MIT License',
                   'Natural Language :: English',
                   'Programming Language :: Python :: 2.7',
                   ],
      keywords='form validator, django, laravel, django-laravel-validator',
      author='younger shen',
      author_email='younger.x.shen@gmail.com',
      url='https://github.com/youngershen/django-laravel-validator',
      license='MIT',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=True,
      install_requires=[
          # -*- Extra requirements: -*-
          'Django >= 1.6'
      ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      )
